@@ -1430,7 +1430,6 @@ def proxy_encrypted_obj(encrypted_obj, handler):
         return encrypted_obj
 
 
-# TODO support 2.0-style encryption
 class DecryptedObjectProxy(PdfObject):
 
     def __init__(self, raw_object: PdfObject, handler):
@@ -1517,6 +1516,14 @@ class DecryptedObjectProxy(PdfObject):
     def container_ref(self):
         return self.raw_object.container_ref
 
+    def __eq__(self, other):
+        # NOTE: this will fail if the dictionary contains "un-decryptable"
+        #  descendants! The diff_analysis module is aware of this restriction,
+        #  but you probably shouldn't use this __eq__ method to compare
+        #  arbitrary objects in a PDF file.
+        return isinstance(other, DecryptedObjectProxy) \
+               and other.decrypted == self.decrypted
+
 
 ASN_DT_FORMAT = "D:%Y%m%d%H%M%S"
 
