@@ -463,6 +463,24 @@ class BasePdfFileWriter(PdfHandler):
             self._info = info
         return info
 
+    def set_custom_trailer_entry(self, key: generic.NameObject,
+                                 value: generic.PdfObject):
+        """
+        Set a custom, unmanaged entry in the document trailer or cross-reference
+        stream dictionary.
+
+        .. warning::
+            Calling this method to set an entry that is managed by pyHanko
+            internally (info dictionary, document catalog, etc.) has undefined
+            results.
+
+        :param key:
+            Dictionary key to use in the trailer.
+        :param value:
+            Value to set
+        """
+        raise NotImplementedError
+
     @property
     def document_id(self) -> Tuple[bytes, bytes]:
         id_arr = self._document_id
@@ -516,7 +534,10 @@ class BasePdfFileWriter(PdfHandler):
             self.root['/Extensions'] = extensions = generic.DictionaryObject()
 
         # check if the extension is already registered,
-        cur_ext_value = extensions.get(ext.prefix_name, None)
+        try:
+            cur_ext_value = extensions[ext.prefix_name]
+        except KeyError:
+            cur_ext_value = None
         extension_dicts = ()
         old_ext_multivalued = False
         if isinstance(cur_ext_value, generic.DictionaryObject):
@@ -550,8 +571,7 @@ class BasePdfFileWriter(PdfHandler):
             # TODO is this still appropriate with the new ExtensionRevision
             #  values?
             if lvl == ext.extension_level:
-                old_ext_applies = True
-                replace_old = False
+                return  # nothing to do
             elif ext.compare_by_level:
                 old_ext_applies = lvl >= ext.extension_level
                 replace_old = not old_ext_applies
@@ -1256,6 +1276,7 @@ class PdfFileWriter(BasePdfFileWriter):
             pdf_name('/Producer'): pdf_string(VENDOR)
         })
 
+        self._custom_trailer_entries = {}
         super().__init__(root, info, id_obj, stream_xrefs=stream_xrefs)
 
         if init_page_tree:
@@ -1335,6 +1356,29 @@ class PdfFileWriter(BasePdfFileWriter):
         sh = PubKeySecurityHandler.build_from_certs(recipients)
         self._assign_security_handler(sh)
 
+    def set_custom_trailer_entry(self, key: generic.NameObject,
+                                 value: generic.PdfObject):
+        """
+        Set a custom, unmanaged entry in the document trailer or cross-reference
+        stream dictionary.
+
+        .. warning::
+            Calling this method to set an entry that is managed by pyHanko
+            internally (info dictionary, document catalog, etc.) has undefined
+            results.
+
+        :param key:
+            Dictionary key to use in the trailer.
+        :param value:
+            Value to set
+        """
+        self._custom_trailer_entries[key] = value
+
+    def _populate_trailer(self, trailer):
+        # allow the base implementation to override internally managed entries
+        trailer.update(self._custom_trailer_entries)
+        super()._populate_trailer(trailer)
+
 
 def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
     """
