@@ -18,7 +18,7 @@ from asn1crypto.algos import SignedDigestAlgorithm
 from pyhanko.sign.general import (
     CertificateStore, SimpleCertificateStore,
     SigningError, optimal_pss_params, simple_cms_attribute,
-    get_pyca_cryptography_hash, as_signing_certificate,
+    get_pyca_cryptography_hash, as_signing_certificate_v2,
     load_private_key_from_pemder, load_cert_from_pemder,
     load_certs_from_pemder, _process_pss_params,
     _translate_pyca_cryptography_cert_to_asn1,
@@ -74,9 +74,6 @@ class Signer:
     def __init__(self, prefer_pss=False):
         self.prefer_pss = prefer_pss
 
-    # TODO I guess that in theory, passing digest_algorithm should never
-    #  be necessary. Should review the ASN.1 syntax for certificates once more.
-
     def get_signature_mechanism(self, digest_algorithm):
         """
         Get the signature mechanism for this signer to use.
@@ -87,8 +84,7 @@ class Signer:
         :param digest_algorithm:
             Digest algorithm to use as part of the signature mechanism.
             Only used if a signature mechanism object has to be put together
-            on-the-fly, and the digest algorithm could not be inferred from
-            the signer's certificate.
+            on-the-fly.
         :return:
             A :class:`.SignedDigestAlgorithm` object.
         """
@@ -105,15 +101,25 @@ class Signer:
         algo = self.signing_cert.public_key.algorithm
         params = None
         if algo == 'ec':
-            mech = 'ecdsa'
+            # with ECDSA, RFC 5753 requires us to encode the digest
+            # algorithm together with the signing algorithm.
+            # The correspondence with the digestAlgorithm field in CMS is
+            # verified separately.
+            if digest_algorithm is None:
+                raise ValueError(
+                    "Digest algorithm required for ECDSA"
+                )
+            mech = digest_algorithm + '_ecdsa'
         elif algo == 'rsa':
             if self.prefer_pss:
                 mech = 'rsassa_pss'
                 if digest_algorithm is None:
-                    raise ValueError("Digest algorithm required")
-                params = optimal_pss_params(
-                    self.signing_cert, digest_algorithm
-                )
+                    raise ValueError(
+                        "Digest algorithm required for RSASSA-PSS"
+                    )
+                params = optimal_pss_params(self.signing_cert, digest_algorithm)
+            elif digest_algorithm is not None:
+                mech = digest_algorithm + '_rsa'
             else:
                 mech = 'rsassa_pkcs1v15'
         else:  # pragma: nocover
@@ -247,8 +253,8 @@ class Signer:
             simple_cms_attribute('message_digest', data_digest),
             # required by PAdES
             simple_cms_attribute(
-                'signing_certificate',
-                as_signing_certificate(self.signing_cert)
+                'signing_certificate_v2',
+                as_signing_certificate_v2(self.signing_cert)
             )
         ]
 
@@ -730,7 +736,8 @@ class SimpleSigner(Signer):
             return None
 
     @classmethod
-    def load_pkcs12(cls, pfx_file, ca_chain_files=None, passphrase=None,
+    def load_pkcs12(cls, pfx_file, ca_chain_files=None,
+                    other_certs=None, passphrase=None,
                     signature_mechanism=None, prefer_pss=False):
         """
         Load certificates and key material from a PCKS#12 archive
@@ -741,6 +748,9 @@ class SimpleSigner(Signer):
         :param ca_chain_files:
             Path to (PEM/DER) files containing other relevant certificates
             not included in the PKCS#12 file.
+        :param other_certs:
+            Other relevant certificates, specified as a list of
+            :class:`.asn1crypto.x509.Certificate` objects.
         :param passphrase:
             Passphrase to decrypt the PKCS#12 archive, if required.
         :param signature_mechanism:
@@ -765,18 +775,26 @@ class SimpleSigner(Signer):
             if ca_chain_files else set()
         if ca_chain is None:  # pragma: nocover
             return None
-
-        (private_key, cert, other_certs) = pkcs12.load_key_and_certificates(
-            pfx_bytes, passphrase
-        )
+        try:
+            (private_key, cert, other_certs_pkcs12) \
+                = pkcs12.load_key_and_certificates(pfx_bytes, passphrase)
+        except (IOError, ValueError, TypeError) as e:
+            logger.error(
+                'Could not load key material from PKCS#12 file', exc_info=e
+            )
+            return None
         kinfo = _translate_pyca_cryptography_key_to_asn1(private_key)
         cert = _translate_pyca_cryptography_cert_to_asn1(cert)
-        other_certs = set(
-            map(_translate_pyca_cryptography_cert_to_asn1, other_certs)
-        )
+        other_certs_pkcs12 = set(map(
+            _translate_pyca_cryptography_cert_to_asn1,
+            other_certs_pkcs12
+        ))
 
         cs = SimpleCertificateStore()
-        cs.register_multiple(ca_chain | set(other_certs))
+        certs_to_register = ca_chain | other_certs_pkcs12
+        if other_certs is not None:
+            certs_to_register |= set(other_certs)
+        cs.register_multiple(certs_to_register)
         return SimpleSigner(
             signing_key=kinfo, signing_cert=cert,
             cert_registry=cs, signature_mechanism=signature_mechanism,
@@ -816,7 +834,7 @@ class SimpleSigner(Signer):
                 key_file, passphrase=key_passphrase
             )
             signing_cert = load_cert_from_pemder(cert_file)
-        except (IOError, ValueError, TypeError) as e:  # pragma: nocover
+        except (IOError, ValueError, TypeError) as e:
             logger.error('Could not load cryptographic material', exc_info=e)
             return None
 
