@@ -8,7 +8,8 @@ from enum import Enum, unique
 from typing import TypeVar, Type, Optional, Union, Iterator, IO
 
 from asn1crypto import (
-    cms, tsp, ocsp as asn1_ocsp, pdf as asn1_pdf, crl as asn1_crl, x509,
+    cms, tsp, ocsp as asn1_ocsp, pdf as asn1_pdf, crl as asn1_crl, x509, keys,
+    core
 )
 from asn1crypto.x509 import Certificate
 from cryptography.hazmat.primitives import hashes
@@ -34,11 +35,12 @@ from .fields import (
     SigSeedValFlags, SigSeedSubFilter
 )
 from .general import (
-    SignatureStatus, find_cms_attribute,
+    SignatureStatus, find_unique_cms_attribute,
     UnacceptableSignerError, KeyUsageConstraints,
     SignatureValidationError,
     validate_sig_integrity, DEFAULT_WEAK_HASH_ALGORITHMS,
-    get_pyca_cryptography_hash, extract_message_digest,
+    get_pyca_cryptography_hash, extract_message_digest, match_issuer_serial,
+    MultivaluedAttributeError, NonexistentAttributeError
 )
 from .timestamps import TimestampSignatureStatus
 
@@ -56,6 +58,8 @@ __all__ = [
     'ValidationInfoReadingError', 'SigSeedValueValidationError'
 ]
 
+from ..pdf_utils.writer import BasePdfFileWriter
+
 logger = logging.getLogger(__name__)
 
 
@@ -82,9 +86,7 @@ class SigSeedValueValidationError(SignatureValidationError):
 
 def _get_signer_predicate(sid: cms.SignerIdentifier):
     if sid.name == 'issuer_and_serial_number':
-        issuer = sid.chosen['issuer']
-        serial_number = sid.chosen['serial_number'].native
-        return lambda c: c.issuer == issuer and c.serial_number == serial_number
+        return lambda c: match_issuer_serial(sid.chosen, c)
     elif sid.name == 'subject_key_identifier':
         # subject key identifier (not all certs have this, but that shouldn't
         # be an issue in this case)
@@ -135,9 +137,9 @@ def _extract_self_reported_ts(signer_info: cms.SignerInfo) \
         -> Optional[datetime]:
     try:
         sa = signer_info['signed_attrs']
-        st = find_cms_attribute(sa, 'signing_time')[0]
+        st = find_unique_cms_attribute(sa, 'signing_time')
         return st.native
-    except KeyError:
+    except (NonexistentAttributeError, MultivaluedAttributeError):
         pass
 
 
@@ -145,13 +147,13 @@ def _extract_tst_data(signer_info, signed=False) -> Optional[cms.SignedData]:
     try:
         if signed:
             sa = signer_info['signed_attrs']
-            tst = find_cms_attribute(sa, 'content_time_stamp')[0]
+            tst = find_unique_cms_attribute(sa, 'content_time_stamp')
         else:
             ua = signer_info['unsigned_attrs']
-            tst = find_cms_attribute(ua, 'signature_time_stamp_token')[0]
+            tst = find_unique_cms_attribute(ua, 'signature_time_stamp_token')
         tst_signed_data = tst['content']
         return tst_signed_data
-    except KeyError:
+    except (NonexistentAttributeError, MultivaluedAttributeError):
         pass
 
 
@@ -419,8 +421,21 @@ class StandardCMSSignatureStatus(SignatureStatus):
 
         validity_info = (
             "The signature is cryptographically "
-            f"{'' if self.intact and self.valid else 'un'}sound."
+            f"{'' if self.intact and self.valid else 'un'}sound.\n\n"
+            f"The digest algorithm used was '{self.md_algorithm}'.\n"
+            f"The signature mechanism used was "
+            f"'{self.pkcs7_signature_mechanism}'."
         )
+        if 'ecdsa' in self.pkcs7_signature_mechanism:
+            ec_params: keys.ECDomainParameters = \
+                cert.public_key['algorithm']['parameters']
+            if ec_params.name == 'named':
+                curve_oid: core.ObjectIdentifier = ec_params.chosen
+                validity_info += (
+                    f"\nThe elliptic curve used for the signer's ECDSA "
+                    f"public key was '{curve_oid.native}' "
+                    f"(OID: {curve_oid.dotted})."
+                )
 
         timing_infos = []
         reported_ts = self.signer_reported_dt
@@ -1843,12 +1858,12 @@ def validate_pdf_ltv_signature(embedded_sig: EmbeddedPdfSignature,
 
 def retrieve_adobe_revocation_info(signer_info: cms.SignerInfo):
     try:
-        revinfo: asn1_pdf.RevocationInfoArchival = find_cms_attribute(
+        revinfo: asn1_pdf.RevocationInfoArchival = find_unique_cms_attribute(
             signer_info['signed_attrs'], "adobe_revocation_info_archival"
-        )[0]
-    except KeyError as e:
+        )
+    except (NonexistentAttributeError, MultivaluedAttributeError) as e:
         raise ValidationInfoReadingError(
-            "No revocation info archival attribute found"
+            "No revocation info archival attribute found, or multiple present"
         ) from e
 
     ocsps = list(revinfo['ocsp'] or ())
@@ -2020,7 +2035,7 @@ def collect_validation_info(embedded_sig: EmbeddedPdfSignature,
 def add_validation_info(embedded_sig: EmbeddedPdfSignature,
                         validation_context: ValidationContext,
                         skip_timestamp=False, add_vri_entry=True,
-                        in_place=False, output=None,
+                        in_place=False, output=None, force_write=False,
                         chunk_size=DEFAULT_CHUNK_SIZE):
     """
     Add validation info (CRLs, OCSP responses, extra certificates) for a
@@ -2048,40 +2063,54 @@ def add_validation_info(embedded_sig: EmbeddedPdfSignature,
     :param chunk_size:
         Chunk size parameter to use when copying output to a new stream
         (irrelevant if ``in_place`` is ``True``).
+    :param force_write:
+        Force a new revision to be written, even if not necessary (i.e.
+        when all data in the validation context is already present in the DSS).
     :return:
         The (file-like) output object to which the result was written.
     """
 
     reader: PdfFileReader = embedded_sig.reader
+    # Take care of this first, so we get any errors re: stream properties
+    # out of the way before doing the (potentially) expensive validation
+    # operations
     if in_place:
-        output = reader.stream
-    # Take care of this first, so we get any errors re: stream properties out
-    # of the way before doing the (potentially) expensive validation operations
-    output = misc.prepare_rw_output_stream(output)
-
-    # if the output is not the same as the input reader's stream, copy the
-    # original file contents to the output before calling add_dss
-    if not in_place:
-        temp_buffer = bytearray(chunk_size)
-        reader.stream.seek(0)
-        misc.chunked_write(temp_buffer, reader.stream, output)
+        working_output = output = reader.stream
+        # Note: random-access I/O is checked at this point.
+        # At the time of writing, this means that all this assertion does is
+        # raise an error if the output is not writable.
+        # We put it in for forwards compatibility & consistency with future
+        # changes to I/O internals.
+        misc.assert_writable_and_random_access(output)
+    else:
+        working_output = misc.prepare_rw_output_stream(output)
 
     paths = collect_validation_info(
         embedded_sig, validation_context, skip_timestamp=skip_timestamp
     )
 
-    # TODO Since add_dss has to re-parse the xref table, this is suboptimal
-    #  in terms of efficiency, but we can iterate on that later
     if add_vri_entry:
         sig_contents = embedded_sig.pkcs7_content.hex().encode('ascii')
     else:
         sig_contents = None
 
-    DocumentSecurityStore.add_dss(
-        output, sig_contents, validation_context=validation_context,
+    pdf_out = IncrementalPdfFileWriter.from_reader(reader)
+    pdf_out.IO_CHUNK_SIZE = chunk_size
+    resulting_dss = DocumentSecurityStore.supply_dss_in_writer(
+        pdf_out, sig_contents, validation_context=validation_context,
         paths=paths
     )
-    return output
+    if force_write or resulting_dss.modified:
+        if in_place:
+            pdf_out.write_in_place()
+        else:
+            pdf_out.write(working_output)
+    elif not in_place:
+        # not in place, but we don't have anything to add -> copy the input
+        # buffer
+        reader.stream.seek(0)
+        misc.chunked_write(bytearray(chunk_size), reader.stream, working_output)
+    return misc.finalise_output(output, working_output)
 
 
 class DocumentSecurityStore:
@@ -2089,7 +2118,8 @@ class DocumentSecurityStore:
     Representation of a DSS in Python.
     """
 
-    def __init__(self, writer, certs=None, ocsps=None, crls=None,
+    def __init__(self, writer: BasePdfFileWriter,
+                 certs=None, ocsps=None, crls=None,
                  vri_entries=None, backing_pdf_object=None):
         self.vri_entries = vri_entries if vri_entries is not None else {}
         self.certs = certs if certs is not None else {}
@@ -2113,6 +2143,17 @@ class DocumentSecurityStore:
             crl_bytes = crl_ref.get_object().data
             crls_seen[crl_bytes] = crl_ref
         self._crls_seen = crls_seen
+        self._modified = False
+
+    @property
+    def modified(self):
+        return self._modified
+
+    def _mark_modified(self):
+        if not self._modified:
+            self._modified = True
+            if self.backing_pdf_object is not None:
+                self.writer.update_container(self.backing_pdf_object)
 
     def _cms_objects_to_streams(self, objs, seen, dest):
         for obj in objs:
@@ -2123,6 +2164,7 @@ class DocumentSecurityStore:
                 ref = self.writer.add_object(
                     generic.StreamObject(stream_data=obj_bytes)
                 )
+                self._mark_modified()
                 seen[obj_bytes] = ref
                 dest.append(ref)
                 yield ref
@@ -2146,6 +2188,7 @@ class DocumentSecurityStore:
         ref = self.writer.add_object(
             generic.StreamObject(stream_data=cert.dump())
         )
+        self._mark_modified()
         self.certs[cert.issuer_serial] = ref
         return ref
 
@@ -2205,11 +2248,14 @@ class DocumentSecurityStore:
         #  we should take that into account
         cert_refs.update(set(self._embed_certs_from_ocsp(ocsps)))
 
+        # TODO do a better job of determining whether the VRI dictionary even
+        #  needs updating.
         if identifier is not None:
             vri = VRI(certs=cert_refs, ocsps=ocsp_refs, crls=crl_refs)
             self.vri_entries[identifier] = self.writer.add_object(
                 vri.as_pdf_object()
             )
+            self._mark_modified()
 
     def as_pdf_object(self):
         """
@@ -2286,12 +2332,10 @@ class DocumentSecurityStore:
             DSS.
         """
         try:
-            dss_ref = handler.root.raw_get(pdf_name('/DSS'))
+            dss_dict = handler.root['/DSS']
         except KeyError as e:
             raise NoDSSFoundError() from e
 
-        dss_dict = dss_ref.get_object()
-
         cert_refs = {}
         cert_ref_list = get_and_apply(dss_dict, '/Certs', list, default=())
         for cert_ref in cert_ref_list:
@@ -2320,9 +2364,8 @@ class DocumentSecurityStore:
             vri_entries = None
 
         # if the handler is a writer, the DSS will support updates
-        if isinstance(handler, IncrementalPdfFileWriter):
+        if isinstance(handler, BasePdfFileWriter):
             writer = handler
-            writer.mark_update(dss_ref)
         else:
             writer = None
 
@@ -2336,19 +2379,20 @@ class DocumentSecurityStore:
         return dss
 
     @classmethod
-    def add_dss(cls, output_stream, sig_contents, *, certs=None,
-                ocsps=None, crls=None, paths=None, validation_context=None):
+    def supply_dss_in_writer(cls, pdf_out: BasePdfFileWriter,
+                             sig_contents, *, certs=None,
+                             ocsps=None, crls=None, paths=None,
+                             validation_context=None) \
+            -> 'DocumentSecurityStore':
         """
         Add or update a DSS, and optionally associate the new information with a
         VRI entry tied to a signature object.
 
-        The result is applied to the output stream as an incremental update.
-
         You can either specify the CMS objects to include directly, or
         pass them in as output from `pyhanko_certvalidator`.
 
-        :param output_stream:
-            Output stream to write to.
+        :param pdf_out:
+            PDF writer to write to.
         :param sig_contents:
             Contents of the new signature (used to compute the VRI hash), as
             as a hexadecimal string, including any padding.
@@ -2365,20 +2409,20 @@ class DocumentSecurityStore:
             to the DSS.
         :param validation_context:
             Validation context from which to draw OCSP responses and CRLs.
+        :return:
+            a :class:`DocumentSecurityStore` object containing both the new
+            and existing contents of the DSS (if any).
         """
-        writer = IncrementalPdfFileWriter(output_stream)
-
         try:
-            dss = cls.read_dss(writer)
+            dss = cls.read_dss(pdf_out)
             created = False
         except ValidationInfoReadingError:
             created = True
-            dss = cls(writer=writer)
+            dss = cls(writer=pdf_out)
 
         if sig_contents is not None:
-            identifier = DocumentSecurityStore.sig_content_identifier(
-                sig_contents
-            )
+            identifier = \
+                DocumentSecurityStore.sig_content_identifier(sig_contents)
         else:
             identifier = None
 
@@ -2406,7 +2450,45 @@ class DocumentSecurityStore:
         # if we're adding a fresh DSS, we need to register it.
 
         if created:
-            dss_ref = writer.add_object(dss_dict)
-            writer.root[pdf_name('/DSS')] = dss_ref
-            writer.update_root()
-        writer.write_in_place()
+            dss_ref = pdf_out.add_object(dss_dict)
+            pdf_out.root[pdf_name('/DSS')] = dss_ref
+            pdf_out.update_root()
+        return dss
+
+    @classmethod
+    def add_dss(cls, output_stream, sig_contents, *, certs=None,
+                ocsps=None, crls=None, paths=None, validation_context=None,
+                force_write=False):
+        """
+        Wrapper around :meth:`supply_dss_in_writer`.
+
+        The result is applied to the output stream as an incremental update.
+
+        :param output_stream:
+            Output stream to write to.
+        :param sig_contents:
+            Contents of the new signature (used to compute the VRI hash), as
+            as a hexadecimal string, including any padding.
+            If ``None``, the information will not be added to any VRI
+            dictionary.
+        :param certs:
+            Certificates to include in the VRI entry.
+        :param ocsps:
+            OCSP responses to include in the VRI entry.
+        :param crls:
+            CRLs to include in the VRI entry.
+        :param paths:
+            Validation paths that have been established, and need to be added
+            to the DSS.
+        :param force_write:
+            Force a write even if the DSS doesn't have any new content.
+        :param validation_context:
+            Validation context from which to draw OCSP responses and CRLs.
+        """
+        pdf_out = IncrementalPdfFileWriter(output_stream)
+        dss = cls.supply_dss_in_writer(
+            pdf_out, sig_contents, certs=certs, ocsps=ocsps,
+            crls=crls, paths=paths, validation_context=validation_context
+        )
+        if force_write or dss.modified:
+            pdf_out.write_in_place()
